from BroachRpc.common import CommonDef,GlobalVariable
from BroachRpc.exception import DefException
from BroachRpc.network import RpcServer,RpcClient
import threading
import logging
import json
import time

"""
rpc 应用通信协议 json
{
  "funcId": "" , 远程调用函数id
  "args": "", 参数
  "kwargs": "", 参数
  "reqId": "0 18", 请求报文id 响应报文的话就是0
  "orgId": "0 18", 响应原报文id 请求报文的话就是0
  "status": "", 处理状态 200 成功 400 报错 404 找不到函数
  "result": "" 远程调用返回值
}
"""

@CommonDef.singleton
class RpcService:
    def __init__(self):
        self.rpcClient = RpcClient.RpcClient()
        self.rpcServer = RpcServer.RpcServer(self.rpcClient)
        self.rpcServer.addNESCallBack(self.__NESRevHandler)
        self.rpcServer.addCIMCallBack(self.__CIMRevHandler)
        self.__orgNES = {}
        self.__orgNESLock = threading.RLock()
        self.registerServer={}
    def serverStart(self):
        self.rpcServer.start()
        GlobalVariable.BroachPool.submit(self.__searchCluster)

    """
    rpc通信调用方法
    """
    def sendRpc(self, ip: str, port: int, funcId, *args, **kwargs):
        #走底层通信NES消息 发送NES消息 对方应答NES消息
        reqId = CommonDef.getId(18)
        rpcTimeOut = int(GlobalVariable.params["rpcTimeOut"])
        with self.__orgNESLock:
            while self.__orgNES.__contains__(reqId):
                reqId = CommonDef.getId(18)
            lock = threading.Condition()
            self.__orgNES[reqId] = {"lock": lock, "result": DefException.CallRpcTimeOutError("响应超时")}
        reqJson = {"funcId": funcId, "args":args, "kwargs": kwargs, "reqId": reqId, "orgId":"0"}
        lock.acquire()
        self.__sendNES(json.dumps(reqJson, ensure_ascii=False), ip, port)
        lock.wait(rpcTimeOut)
        lock.release()

        result = self.__orgNES.get(reqId).get("result")
        with self.__orgNESLock:
            self.__orgNES.pop(reqId)
        if isinstance(result, DefException.CallRpcTimeOutError) is True:
            raise result
        else:
            status = result.get("status")
        if status == "400":
            raise DefException.RpcError(result.get("result"))
        elif status == "404":
            raise DefException.RpcFuncNotFundError(result.get("result"))
        else:
            return result.get("result")
    """
    发送NES消息
    """
    def __sendNES(self, reqJsonStr, ip, port):
        msgId = CommonDef.getId(18)
        sendResult = self.rpcClient.sendNES(msgId, reqJsonStr, ip, port)
        while sendResult == "idRepeat":
            logging.error("NES通信消息id，重复异常。重新随机id")
            msgId = CommonDef.getId(18)
            sendResult = self.rpcClient.sendNES(msgId, reqJsonStr, ip, port)
        logging.debug("发送NES消息: "+reqJsonStr)
    """
    微服务上线规则 非udp广播模式
    先去配置的ip列表里请求集群中的主机获取集群信息 然后注册本机和本机的路由给所有集群里的主机
    """
    def __searchCluster(self):
        addressList = GlobalVariable.params.get("clusterAddress")
        localIp = str(GlobalVariable.params.get("rpcIp"))
        localPort = str(GlobalVariable.params.get("rpcPort"))

        time.sleep(3) #等待一会儿 udp服务器起来再执行
        #获取集群实例总列表
        for address in addressList:
            ip = address.split(":")[0]
            port = address.split(":")[1]
            self.registerServer[address] = 0
            serverInstanceList = None
            try:
                #获取其他主机里的集群主机信息
                serverInstanceList = self.sendRpc(ip, port, "__getServerInstance")
            except Exception as e:
                logging.error(str(e))
            if serverInstanceList is not None:
                for item in serverInstanceList:
                    if item not in self.registerServer.keys():
                        self.registerServer[item] = 0
        #通知所有实例 本机上线
        for item in self.registerServer.keys():
            ip = item.split(":")[0]
            port = item.split(":")[1]
            msg = localIp+":"+localPort+":"+"online"
            self.rpcClient.sendCIM(msg, ip, port)
        time.sleep(3)
        logging.debug("启动集群存活探测线程")
        isSetLocalRoute = []
        while True:
            for item in self.registerServer.keys():
                ip = item.split(":")[0]
                port = item.split(":")[1]
                heartbeat = self.registerServer.get(item)
                if heartbeat == 0:
                    logging.debug("发送上线事件 -> " + ip + ":" + port)
                    msg = localIp + ":" + localPort + ":" + "online"
                    self.rpcClient.sendCIM(msg, ip, port)
                else:
                    if item not in isSetLocalRoute:
                        try:
                            self.sendRpc(ip, port, "__setRpcRoute",
                                         {"address": GlobalVariable.params.get("rpcIp") + ":"
                                                     + str(GlobalVariable.params.get("rpcPort")),
                                          "route": getLocalRoute()})
                            isSetLocalRoute.append(item)
                        except Exception:
                            logging.exception(exc_info=True, msg="向远程主机注册本机路由表失败 ->"+ ip + ":" + port)
                    msg = localIp + ":" + localPort + ":" + "heartbeat"
                    self.rpcClient.sendCIM(msg, ip, port)
            time.sleep(3)

    """
    处理收到的CIM消息
    """
    def __CIMRevHandler(self, msg):
        logging.debug("收到CIM消息: "+ msg)
        ip = msg.split(":")[0]
        port = msg.split(":")[1]
        thing = msg.split(":")[2] #cim事件
        if thing == "online":
            #主机上线事件
            address = ip+":"+port
            if address not in self.registerServer.keys():
                logging.debug("收到主机上线事件 -> " + ip+":"+port)
                # 发送本机路由表给上线的主机
                try:
                    self.rpcClient.sendCIM(GlobalVariable.params.get("rpcIp") + ":" + str(GlobalVariable.params.get("rpcPort"))+":heartbeat", ip, port)
                    self.sendRpc(ip, port, "__setRpcRoute",
                                 {"address": GlobalVariable.params.get("rpcIp") + ":" + str(GlobalVariable.params.get("rpcPort")),
                                  "route": getLocalRoute()})

                    self.registerServer[address] = time.time()
                except Exception as e:
                    logging.exception(exc_info=True)
                    logging.error("远程主机注册本机路由表失败 " + address + " "+ str(e))
        elif thing == "heartbeat":
            #心跳事件
            self.registerServer[ip+":"+port] = time.time()


    """
    处理收到的NES消息
    """
    def __NESRevHandler(self, msg, revIp, revPort):
        try:
            revJson = json.loads(msg)
        except Exception as e:
            logging.exception(exc_info=True, msg="json 解析异常 "+str(e))
            return
        logging.debug("接收到NES消息: "+ msg)
        orgId = revJson.get("orgId")
        if str(orgId) != "0":
            #响应回调 结束阻塞
            if self.__orgNES.get(orgId) is not None:
                with self.__orgNESLock:
                    self.__orgNES.get(orgId)["result"] = revJson
                lock = self.__orgNES.get(orgId).get("lock")
                lock.acquire()
                lock.notify()
                lock.release()
        else:
            reqId = revJson.get("reqId")
            funcId = revJson.get("funcId")
            fn = GlobalVariable.FuncRoute.get(funcId)
            if fn is None:
                re = {"status": "404", "result": "funcNotFund", "orgId": reqId}
                self.__sendNES(json.dumps(re), revIp, revPort)
            else:
                args = revJson.get("args")
                kwargs = revJson.get("kwargs")
                try:
                    result = fn(*args, **kwargs)
                    status = "200"
                except Exception as e:
                    result = str(e)
                    status = "400"
                re = {"status": status, "result": result, "orgId": reqId}
                try:
                    reJson = json.dumps(re)
                except Exception as e:
                    reJson = json.dumps({"status": "400", "result": str(e), "orgId": reqId})
                self.__sendNES(reJson, revIp, revPort)

"""
返回本机信息
"""
def getServerInstance():
    if instance is not None:
        return list(instance.registerServer.keys())
    else:
        return []

"""
外部主机注册路由表到本机
"""
def setRpcRoute(rpcRoute:dict):
    if rpcRoute is not None:
        address = rpcRoute.get("address")
        route = rpcRoute.get("route")
        for funcId in route:
            funcIdInfo = GlobalVariable.FuncRouteRpc.get(funcId)
            if funcIdInfo is None:
                GlobalVariable.FuncRouteRpc[funcId] = {address:{"errorRatio": 0, "reqNum": 0}}
            else:
                if funcIdInfo.get(address) is None:
                    funcIdInfo[address] = {"errorRatio": 0, "reqNum": 0}
    return "ok"

"""
获取本机可暴露的路由list
"""
def getLocalRoute():
    re = []
    for item in GlobalVariable.FuncRoute.keys():
        if item[0:2] != "__":
            re.append(item)
    return re
GlobalVariable.FuncRoute["__getServerInstance"] = getServerInstance
GlobalVariable.FuncRoute["__setRpcRoute"] = setRpcRoute
instance = None



from BroachRpc.common import GlobalVariable
from BroachRpc.exception import DefException
import logging
import copy
from BroachRpc.network import RpcService

"""
rpc装饰器 获取程序启动时的函数路由表
"""
def rpcRoute(name=None):
    def getFun(fn):
        if name is None:
            routeName = fn.__name__
        else:
            routeName = name
        if GlobalVariable.FuncRoute.get(routeName) is None:
            GlobalVariable.FuncRoute[routeName] = fn
            logging.debug("function: "+routeName+" registry route success")
        else:
            logging.error(routeName+" registry route fail，name is replace")
        return fn
    return getFun

"""
rpc装饰器 调用rpc服务
"""
def rpcCall(name=None):
    def getFun(fn):
        def getParams(*args, **kwargs):
            if name is None:
                routeName = fn.__name__
            else:
                routeName = name
            if routeName in GlobalVariable.FuncRoute.keys():
                logging.info("调用内部函数 -> "+routeName)
                rpcFn = GlobalVariable.FuncRoute.get(routeName)
                return rpcFn(*args, **kwargs)
            else:
                if routeName in GlobalVariable.FuncRouteRpc.keys():
                    rpcFnInfo = GlobalVariable.FuncRouteRpc.get(routeName)
                    if rpcFnInfo is not None:
                        ipList = rpcFnInfo.keys()
                        reqAddress = ""
                        minReqNum = 100
                        for address in ipList:
                            info = rpcFnInfo.get(address)
                            reqNum = info.get("reqNum")
                            if reqNum >= 100:
                                info["reqNum"] = 0
                                reqAddress = address
                                break
                            if reqNum < minReqNum:
                                minReqNum = reqNum
                                reqAddress = address
                        if reqAddress == "":
                            raise DefException.RpcFuncNotFundError(routeName+" 没有可用的服务端")
                        rpcFnInfo[reqAddress]["reqNum"] = minReqNum + 1
                        ip = reqAddress.split(":")[0]
                        port = reqAddress.split(":")[1]
                        return RpcService.instance.sendRpc(ip,port,routeName, *args, **kwargs)
                    else:
                        raise DefException.RpcFuncNotFundError(routeName+" is not Fund")
                else:
                    raise DefException.RpcFuncNotFundError(routeName+" is not Fund")
            #logging.debug("rpc调用"+routeName)
            #re = RpcService.RpcService.sendRpc(routeName, )
            #logging.debug("rpc调用结束 结果:"+re)
            #return re
        return getParams
    return getFun


import threading
import logging
import time
from exception.DefException import QueueOverflowError, ThreadPoolIsShutdown

class Pool:
    def __init__(self, core=1, threadMax=5, keepAliveTime=6, queue=None, poolName="pool", taskTimeOut=10):
        """
        :param core:  核心线程数
        :param max:  最大线程数
        :param sleepTime: 线程闲置回收时间
        :param queue: 队列对象
        :param poolName: 线程名
        :param taskTimeOut 任务执行返回值超时时间 即超过这个时间 异步返回值会提示timeout error
        """
        self.core = core
        self.threadMax = threadMax
        self.keepAliveTime = keepAliveTime
        self.queue = queue
        if queue is None:
            self.queue = ArrayQueue()
        if isinstance(queue, Queue) is False:
            raise Exception("请传入Queue类的实例对象")
        self.threadNum = 0
        self.poolLock = threading.RLock()
        self.poolName = poolName
        self.taskTimeOut = taskTimeOut
        self.poolStatus = True
        for i in range(0, self.core):
            self.createThread(True)
        logging.debug("线程池 "+self.poolName+" 创建完成")

    def createThread(self, isCore=False):
        """
        创建线程 false标识创建失败
        :param isCore: 是否核心线程
        :return: 成功或者失败
        """
        with self.poolLock:
            if self.threadNum >= self.threadMax:
                return False
            else:
                thread = threading.Thread(target=self.__threadFunc, kwargs={"isCore": isCore})
                threadName = self.poolName + "-" + str(self.threadNum + 1)
                thread.setName(threadName)
                self.threadNum = self.threadNum + 1
                thread.start()
                return True

    def __threadFunc(self, isCore):
        """
        工作线程执行函数
        会像队列的get方法获取任务，None标识退出线程
        或者poolStatus = false 退出循环结束工作线程
        :param isCore: 是否核心线程
        :return: void
        """
        data = {"isCore": isCore, "lock": threading.Condition(), "isWork": False}
        while self.poolStatus:
            task = self.queue.get(data, self)
            if task is None:
                break
            data["isWork"] = True
            try:
                task["return"] = task.get("func")(*task.get("args"), **task.get("kwargs"))
            except Exception as e:
                task["return"] = "taskError: "+str(e)
                logging.exception("任务执行异常 " + str(task), exc_info=True)
            data["isWork"] = False
        logging.debug("exit")
        with self.poolLock:
            self.threadNum = self.threadNum - 1

    def submit(self, fn, *args, **kwargs):
        """
        提交任务
        {"func": fn, "args": args, "kwargs": kwargs, "return": None}
        :param fn: 执行函数对象
        :param args: 参数
        :param kwargs: 参数
        :return: 异步对象dict
        """
        if self.poolStatus:
            task = {"func": fn, "args": args, "kwargs": kwargs, "return": None}
            return self.queue.put(task, self)
        else:
            raise ThreadPoolIsShutdown("线程池已关闭，无法提交任务")

    def shutDown(self):
        """
        关闭线程池
        :return:
        """
        self.poolStatus = False

class Queue:
    def put(self, task:dict, pool:Pool):
        """
        线程池调用put方法提交任务
        :param task: {"func": fn, "args": args, "kwargs": kwargs, "return": None}
        :param pool: 线程池对象
        :return: {"func": fn, "args": args, "kwargs": kwargs, "return": None}
        """
        raise Exception("请实现队列的PUT方法")
    def get(self, thread:dict, pool:Pool):
        """
        线程池调取get方法获取任务
        :param thread: {"isCore": isCore, "lock": threading.Condition(), "isWork": False}
        :param pool: 线程池对象
        :return: void
        """
        raise Exception("请实现队列的GET方法")

class ArrayQueue(Queue):
    def __init__(self, queueMax=5, isExcInfo=False, createThreadThreshold=None):
        """
        :param queueMax: 队列最大值
        :param isExcInfo: 溢出后是否报错 默认不报错交给提交任务的线程执行任务
        :param createThreadThreshold: 队列满足阈值后创建非核心线程
        :param queueFullWaitTime: 队列满了以后 等待任务接取时间
        """
        self.queueMax = queueMax
        self.queueList = []
        self.queueListLock = threading.RLock()
        self.threadList = []
        self.threadListLock = threading.RLock()
        self.isExcInfo = isExcInfo
        self.createThreadThreshold = createThreadThreshold
        if createThreadThreshold is None:
            self.createThreadThreshold = queueMax

    def put(self, task:dict, pool:Pool):
        """
        当提交任务到线程池时 线程池会调用队列的put方法传递任务对象处理 并将put方法的返回值作为异步对象返回给提交任务的线程
        队列满了以后 根据设置 是否抛出异常 中断线程
        :param task: {"func": fn, "args": args, "kwargs": kwargs, "return": None}
        :param pool: 调用的线程池对象
        :return: {"func": fn, "args": args, "kwargs": kwargs, "return": wait} dict 待任务执行完成后 return键值将会获得返回对象
        """
        if len(self.queueList) >= self.queueMax and pool.threadNum == pool.threadMax:
            if self.isExcInfo:
                raise QueueOverflowError("任务队列已满")
            else:
                logging.error("任务队列已满，将交由提交线程执行任务")
                try:
                    task["return"] = task.get("func")(*task.get("args"), **task.get("kwargs"))
                    return task
                except Exception as e:
                    task["return"] = "taskError: "+str(e)
                    logging.exception("任务执行异常 " + str(task), exc_info=True)

        with self.queueListLock:
            self.queueList.append(task)

        if len(self.queueList) >= self.createThreadThreshold:
            pool.createThread()

        for thread in self.threadList:
            if thread["isWork"] is False:
                lock = thread["lock"]
                lock.acquire()
                lock.notify()
                lock.release()
        return task

    def get(self, thread:dict, pool:Pool):
        """
        工作线程获取任务 调用get方法 从队列中获取任务
        没有获取到任务 就用锁阻塞线程 等待唤醒获取任务
        当非核心线程闲置时间超过阈值时 返回None 结束工作线程
        :param thread: 线程对象 {"isCore": isCore, "lock": threading.Condition(), "isWork": False}
        :param pool: 线程池对象
        :return:
        """
        if thread not in self.threadList:
            with self.threadListLock:
                self.threadList.append(thread)

        while True:
            task = None
            with self.queueListLock:
                if len(self.queueList) > 0:
                    task = self.queueList[0]
                    self.queueList.remove(task)
            if task is not None:
                return task
            else:
                sTime = time.time()
                lock = thread["lock"]
                lock.acquire()
                lock.wait(pool.keepAliveTime)
                lock.release()
                eTime = time.time()
                if thread["isCore"] == False and (eTime - sTime) >= pool.keepAliveTime:
                    with self.threadListLock:
                        self.threadList.remove(thread)
                    return None
                if pool.poolStatus == False and len(self.queueList) == 0:
                    return None

class SyncQueue(Queue):
    def __init__(self):
        self.threadList = []
        self.threadListLock = threading.RLock()
        self.task = {}
        self.taskLock = threading.RLock()
        self.waitTask = []
        self.waitTaskLock = threading.RLock()

    def put(self, task:dict, pool:Pool):
        """
        阻塞提交任务的线程 直到提交的任务被线程获取执行
        :param task: 任务对象 {"isCore": isCore, "lock": threading.Condition(), "isWork": False}
        :param pool: 线程池对象
        :return:
        """
        putLock = threading.Condition()

        idleThread = None
        with self.threadListLock:
            for thread in self.threadList:
                if thread["isWork"] is False:
                    idleThread = thread
                    thread["isWork"] = True
        if idleThread is None:
            putLock.acquire()
            data = {"putLock": putLock, "task": task}
            with self.waitTaskLock:
                self.waitTask.append(data)
            pool.createThread()
            putLock.wait()
            putLock.release()
        else:
            with self.taskLock:
                self.task[idleThread["lock"]] = task
            lock = idleThread["lock"]
            lock.acquire()
            lock.notify()
            lock.release()
        return task

    def get(self, thread:dict, pool:Pool):
        """
        没有获取到任务就阻塞 获取到任务就把原本阻塞提交任务线程的锁释放掉
        :param thread: 线程对象
        :param pool: 线程池对象
        :return:
        """
        lock = thread["lock"]
        if thread not in self.threadList:
            with self.threadListLock:
                self.threadList.append(thread)
                self.task[lock] = None
        while True:
            with self.taskLock:
                task = self.task[lock]
                self.task[lock] = None
            if task is not None:
                return task
            else:
                if len(self.waitTask) != 0:
                    with self.waitTaskLock:
                        data = self.waitTask[0]
                        self.waitTask.remove(data)
                    putLock = data["putLock"]
                    task = data["task"]
                    putLock.acquire()
                    putLock.notify()
                    putLock.release()
                    return task
                sTime = time.time()
                lock.acquire()
                lock.wait(pool.keepAliveTime)
                lock.release()
                eTime = time.time()
                if thread["isCore"] == False and (eTime - sTime) >= pool.keepAliveTime:
                    with self.threadListLock:
                        self.threadList.remove(thread)
                    return None
                if pool.poolStatus is False and self.task is None:
                    return None

#SyncPool = Pool(core=1, threadMax=5, keepAliveTime=6, queue=SyncQueue(), poolName="SyncPool")
#BroachPool = Pool(core=3, threadMax=15, keepAliveTime=9,
#                 queue=ArrayQueue(queueMax=1000, isExcInfo=False, createThreadThreshold=2), poolName="BroachPool")


import threading

#appName
name = "BroachRpc"

#版本
version = "1.0"

#日志级别
logLevel = "info"

#配置
params = {
    "rpcPort": 18081,
    "rpcIp": "127.0.0.1",
    "clusterAddress" : ["127.0.0.1:18080"],
    "rpcTimeOut": 6,
    "udpTimeOut": 3,
    "threadPool" : {
      "core" : 3,
      "threadMax" : 15,
      "keepAliveTime" : 9,
      "queueMax" :  1000,
      "isExcInfo" : "False",
      "createThreadThreshold" : 2
    }
  }

#全局线程池
BroachPool = None

#本机函数路由映射 "__" (双下划线)开头命名的路由表示当前主机私有的映射 不会暴露给外部主机
FuncRoute = {}

#其他机器的路由表 { "funcId": {"ip:port": {"errorRatio": 0, "reqNum": 0, "updateTime": time}, ...} }
FuncRouteRpc = {}
FuncRouteRpcLock = threading.RLock()

#udp缓冲区大小
BufferLen = 3028

#NES类udp消息间隔重传
IntervalUdp = 0.15

#定时任务组 [{ "fn": fn, "timeStep": 6 }]
TimerTasks = []



"""
线程池队列长度溢出异常
"""
class QueueOverflowError(Exception):
    def __init__(self, message):
        self.message = message

"""
接收udp报文缓冲区溢出异常
"""
class ReceiveBuffOverError(Exception):
    def __init__(self, message):
        self.message = message

"""
线程池已停止
"""
class ThreadPoolIsShutdown(Exception):
    def __init__(self, message):
        self.message = message
"""
NES消息发送超时
"""
class RpcSendNESTimeOutError(Exception):
    def __init__(self, message):
        self.message = message
"""
rpc远程调用响应超时
"""
class CallRpcTimeOutError(Exception):
    def __init__(self, message):
        self.message = message
"""
rpc远程调用远程服务端无此函数
"""
class RpcFuncNotFundError(Exception):
    def __init__(self, message):
        self.message = message


"""
rpc熔断异常
"""
class RpcFuncFusingError(Exception):
    def __init__(self, message):
        self.message = message

"""
rpc调用通用异常
"""
class RpcError(Exception):
    def __init__(self, message):
        self.message = message